from fabric import api as fab
from os.path import exists
from contrib.django import DjangoFabric
from contrib.virtualenv import VirtualenvFabric

__all__ = ['BaseFabric', 'DjangoFabric', 'VirtualenvFabric']

class BaseFabric(object):

    hosts = []
    remote_project_path = ''
    local_project_path = ''

    search_dirs = []
    search_exclude_patterns = [".pyc", ".svn", ".tmp_"]

    def __init__(self):
        fab.hosts = self.hosts

    def _remote(self):
        '''
        Returns true if we run fabfile remotely.
        Script check existing local_project_path directory
        '''
        return not exists(self.local_project_path)

    def search(self, string=''):
        '''
        Search string amoung source code inside directories from `search_dirs` property
        '''
        fab.local('grep -r %(string)s %(dirs)s %(exclude)s' % {
            'string': string,
            'dirs': ' '.join(self.search_dirs),
            'exclude': '| ' + ' | '.join(['grep -v "%s"' % pattern for pattern in self.search_exclude_patterns]) if len(self.search_exclude_patterns) else '',
        }, capture=False)

    def del_pyc(self):
        '''
        Remove all .pyc files inside in fabfile and child directories
        '''
        fab.local('find -name \*.pyc -delete')


from fabric.api import lcd, local
from fabric.context_managers import prefix, settings
from contextlib import contextmanager
from os.path import join, isdir, relpath
from os import listdir

class VirtualenvFabric(object):
    '''
    Fabric command file with base virtualenv commands
    '''
    use_virtualenv = True
    virtualenv_dir = 'env'
    patched_applications = [
        # list of applications in format below:
        # applications from repo in env/src/
        # ('django', 'snv'),
        # ('django-mptt-comments', 'git'),
        # applications from pypi in env/lib/python2.6/site-packages/ !!!! exactly app name, not 'django-pagination'
        # ('pagination', 'lib'),
        # TODO: refactor for standard naming, excluding way of installing
    ]
    diff_dir = 'diffs'
    applications_dir = 'apps'

    def _site_packages_path(self):
        return join(self.local_project_path, self.virtualenv_dir, 'lib', 'python2.6', 'site-packages')

    def _site_package_path(self, app_name):
        return join(self._site_packages_path(), app_name)

    def _source_package_path(self, app_name):
        return join(self.local_project_path, self.virtualenv_dir, 'src', app_name)

    @contextmanager
    def virtualenv(self):
        '''
        Context manager for running command in virtualenv environment
        '''
        with prefix('source %s/bin/activate' % join(self.local_project_path, self.virtualenv_dir)):
            yield

    def patch(self):
        '''
        Walk over 'diffs' directory and patch every application
        '''
        vcs_commands = {
            'git': 'git checkout .',
            'svn': 'svn revert -R .',
            'hg': 'hg revert .',
        }
        for patch in listdir(join(self.local_project_path, self.diff_dir)):
            app_name, vcs = patch.split('.')[:-1]
            if vcs == 'lib':
                app_dir = self._site_packages_path()
                if isdir(self._site_package_path(app_name)) and isdir(self._site_package_path('%s.original' % app_name)):
                    # delete patched app and move original app to the normal place
                    with lcd(app_dir):
                        local('rm -R %(app)s && mv %(app)s.original %(app)s' % {'app': app_name}, capture=False)
                # copy app from normal to reserve place for keeping
                with lcd(app_dir):
                    local('cp -R %(app)s %(app)s.original' % {'app': app_name}, capture=False)
            else:
                app_dir = self._source_package_path(app_name)
                # revert files from repo
                with lcd(app_dir):
                    local(vcs_commands[vcs], capture=False)

            with lcd(app_dir):
                patch_command = 'patch -p1' if vcs == 'hg' else 'patch -p0'
                local('%s < %s' % (patch_command, join(self.local_project_path, self.diff_dir, patch)), capture=False)

    def diff_dump(self):
        '''
        Walk over map of patched applications and make diff files in 'diffs' directory
        '''
        vcs_commands = {
            'git': 'git diff --no-prefix',
            'svn': 'svn diff',
            'hg': 'hg diff' #TODO: remove prefixes a/ and b/ in diff files like in git
        }
        for app_name, vcs in self.patched_applications:
            patch_path = join(self.local_project_path, self.diff_dir, '%s.%s.diff' % (app_name, vcs))
            if vcs == 'lib':
                app_dir = self._site_packages_path()
                with lcd(app_dir):
                    with settings(warn_only=True):
                        local('diff -rcx .pyc %s.original %s > %s' % (app_name, app_name, patch_path), capture=False)
            else:
                app_dir = self._source_package_path(app_name)
                with lcd(app_dir):
                    local('%s > %s' % (vcs_commands[vcs], patch_path), capture=False)

    def symlink(self, app_name):
        '''
        Create symlink from 'apps' dir to the site-packages or source application dir
        '''
        source_dir = None
        if isdir(self._source_package_path(app_name)):
            source_dir = self._source_package_path(app_name)
        elif isdir(self._site_package_path(app_name)):
            source_dir = self._site_package_path(app_name)

        if source_dir:
            # convert absolute path to relative from application dir
            source_dir = relpath(source_dir, join(self.local_project_path, self.applications_dir))
            local('ln -s %s %s' % (source_dir, join(self.applications_dir, app_name)))


import inspect
import sys

def add_class_methods_as_module_level_functions_for_fabric(instance, module_name):
    '''
    Utility to take the methods of the instance of a class, instance,
    and add them as functions to a module, module_name, so that Fabric
    can find and call them. Call this at the bottom of a module after
    the class definition. Returns a list of method for __all__ variable,
    otherwise command 'fab -l' will show extra commands.
    '''
    # get the module as an object
    module_obj = sys.modules[module_name]

    # Iterate over the methods of the class and dynamically create a function
    # for each method that calls the method and add it to the current module
    for method in inspect.getmembers(instance, predicate=inspect.ismethod):
        method_name, method_obj = method

        if not method_name.startswith('_'):
            # get the bound method
            func = getattr(instance, method_name)

            # add the function to the current module
            setattr(module_obj, method_name, func)

    return [name for name in dir(instance) if name[0] != '_' and callable(getattr(instance, name))]


from fabric.api import local
import re

class DjangoFabric(object):
    '''
    Fabric command file with base django's manage.py commands
    '''
    managefile_path = './manage.py'
    test_default_app = None
    test_settings = None
    shell_plus = False
    devserver_port = 8000

    def _manage(self, command):
        '''
        Run manage.py command
        '''
        if getattr(self, 'use_virtualenv', False):
            # use virtualenv context manager
            with self.virtualenv():
                local('%(manage)s %(command)s' % {
                    'manage': self.managefile_path,
                    'command': command,
                }, capture=False)
        else:
            local('%(manage)s %(command)s' % {
                'manage': self.managefile_path,
                'command': command,
            }, capture=False)

    def dev(self):
        '''
        Run Django's dev server
        '''
        self._manage('runserver %(host)s:%(port)d' % {
            'host': self.hosts[0] if self._remote() else '127.0.0.1',
            'port': self.devserver_port,
        })

    def sh(self):
        '''
        Run Django's standart shell or shell from django_extentions application if `shell_plus=True`
        '''
        self._manage('shell_plus' if self.shell_plus else 'shell')

    def test(self, test_name=''):
        '''
        Run Django's tests. Argument can be application name, name of test class or method of test class
        '''
        if self.test_default_app:
            # without args => test default application
            if not test_name:
                test_name = self.test_default_app

            # arg is name of test class => add default application
            if re.search(r'^[^\.]+Test', test_name):
                test_name = '.'.join([self.test_default_app, test_name])

        self._manage('test %(name)s %(settings)s' % {
            'name': test_name,
            'settings': ('--settings=%s' % self.test_settings) if self.test_settings else '',
        })